// Copyright (C) 2014 by Karlsruher Institut fuer Technologie (KIT), Institut fuer Mess- und Regelungstechnik
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of arDroneManual/Autonomous.
// 
// arDroneManual/Autonomous is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// arDroneManual/Autonomous is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even 
// the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
// 
// You should have received a copy of the GNU General Public License along with arDroneManual/Autonomous. If not, see 
// <http://www.gnu.org/licenses/>.


// undistortion, rectification and cropping of raw camera images
// =============================================================

#pragma once

#include "hawaii/GPU/autoMat.h"
#include "hawaii/common/hardware.h"
#include "hawaii/GPU/nvccFixHeaderAbove.h"
#include <boost/thread/shared_mutex.hpp>
#include "hawaii/GPU/nvccFixHeaderBelow.h"

namespace toast2 {
namespace stereo {

// camera and projection matrices
template< typename Matx >
struct Intrinsic : public Matx {
	Intrinsic() :
		Matx( Matx::all( NAN ) ) {
	}
	Intrinsic( const typename Matx::value_type focalLengthUArg,
	           const typename Matx::value_type focalLengthVArg,
	           const typename Matx::value_type principalPointUArg,
	           const typename Matx::value_type principalPointVArg ) :
		Matx( Matx::all( 0.0 ) ) {
		(*this)( 0, 0 ) = focalLengthUArg    ;
		(*this)( 1, 1 ) = focalLengthVArg    ;
		(*this)( 0, 2 ) = principalPointUArg ;
		(*this)( 1, 2 ) = principalPointVArg ;
		(*this)( 2, 2 ) = 1.0                ;
	}
	                  Intrinsic(  const Matx& that ) :        Matx(             that ) { }
	Intrinsic< Matx > operator =( const Matx& that ) { return Matx::operator =( that ) ; }
	typename Matx::value_type focalLengthU()    const { return (*this)( 0, 0 ) ; }
	typename Matx::value_type focalLengthV()    const { return (*this)( 1, 1 ) ; }
	typename Matx::value_type principalPointU() const { return (*this)( 0, 2 ) ; }
	typename Matx::value_type principalPointV() const { return (*this)( 1, 2 ) ; }
} ;
typedef Intrinsic< cv::Matx< double, 3, 3 > > Camera     ;
typedef Intrinsic< cv::Matx< double, 3, 4 > > Projection ;

// distortion parameters
struct Distortion : public cv::Vec< double, 5 > {
	Distortion( const double radial1Arg     = NAN,
	            const double radial2Arg     = 0.0,
	            const double tangential1Arg = 0.0,
	            const double tangential2Arg = 0.0,
	            const double radial3Arg     = 0.0 ) :
		cv::Vec< double, 5 >( radial1Arg, radial2Arg, tangential1Arg, tangential2Arg, radial3Arg ) {}
	                     Distortion( const cv::Vec< double, 5 >& that ) :        cv::Vec< double, 5 >(             that ) { }
	cv::Vec< double, 5 > operator =( const cv::Vec< double, 5 >& that ) { return cv::Vec< double, 5 >::operator =( that ) ; }
} ;

// forward declaration
class Rectifier ;

// generic remapping
class Remapper {
	
	// initialize or update based on LUTs from output to input coordinates
	public:
	Remapper(    const cv::Mat mapUArg = cv::Mat(),
	             const cv::Mat mapVArg = cv::Mat() ) ;
	void update( const cv::Mat mapUArg,
	             const cv::Mat mapVArg ) ;
	protected:
	friend class Rectifier ;
	void updateUnlocked( const cv::Mat mapUArg,
	                     const cv::Mat mapVArg,
	                     const cv::Mat mapPixArg,
	                     const cv::Mat mapSubPixArg ) ;
	
	// perform transformation on either CPU or GPU, optionally crop to ROI
	public:
	void CPU( const cv::Mat  src,
	                cv::Mat& dst,
	                cv::Rect roi   = cv::Rect(),
	          const int      cores = hawaii::CPUCores ) const ;
	void GPU( const cv::gpu::GpuMat  src,
	                cv::gpu::GpuMat& dst,
	                cv::Rect         roi    = cv::Rect(),
	                cv::gpu::Stream& stream = cv::gpu::Stream::Null() ) const ;
	
	// get initialization state and output size
	public:
	bool     initialized() const ;
	cv::Size sizeDst()     const ;
	
	// internal data: GPUs directly read copies of the given floating-point maps in "map{U,V}" through texture memory. 
	//                CPUs use the automatically computed faster index-based representation in "map(Sub)Pix". The shared 
	//                mutex allows multiple readers - i.e. concurrent calls to "CPU()" and "GPU()" - or one writer - 
	//                i.e. a call to "update()" - at any given time.
	protected:
	hawaii::AutoMat mapU,
	                mapV ;
	cv::Mat mapPix,
	        mapSubPix ;
	mutable boost::shared_mutex mutex ;
} ;

// undistort images from a mono-camera
class Undistorter : public Remapper {
	
	// create un-initialized instance
	public:
	Undistorter() ;
	
	// initialize or update based on distorted image/camera parameters
	public:
	Undistorter( const Distortion distortion,
	             const Camera     cameraSrc,
	             const cv::Size   sizeSrc,
	             const bool       equalFocalLengths = true ) ;
	void update( const Distortion distortion,
	             const Camera     cameraSrc,
	             const cv::Size   sizeSrc,
	             const bool       equalFocalLengths = true ) ;
	
	// initialize or update based on distorted/undistorted image/camera parameters
	// user note: If "equalFocalLengths" is set, "cameraDstArg" and "sizeDst" will be changed if necessary. 
	public:
	Undistorter( const Distortion distortion,
	             const Camera     cameraSrc,
	             const Camera     cameraDstArg,
	             const cv::Size   sizeDst,
	             const bool       equalFocalLengths = true ) ;
	void update( const Distortion distortion,
	             const Camera     cameraSrc,
	             const Camera     cameraDstArg,
	             const cv::Size   sizeDst,
	             const bool       equalFocalLengths = true ) ;
	
	// initialize or update based on LUTs and undistorted camera matrix
	// user note: If "equalFocalLengths" is set, "cameraDstArg" will be changed if necessary.
	public:
	Undistorter( const cv::Mat mapUArg,
	             const cv::Mat mapVArg,
	             const Camera  cameraDstArg,
	             const bool    equalFocalLengths = true ) ;
	void update( const cv::Mat mapUArg,
	             const cv::Mat mapVArg,
	             const Camera  cameraDstArg,
	             const bool    equalFocalLengths = true ) ;
	
	// get output camera matrix
	public:
	Camera cameraDst() const ;
	protected:
	Camera cameraDstM ;
} ;

// rectify images from a stereo-camera (includes undistortion)
class Rectifier {
	
	// create un-initialized instance
	public:
	Rectifier() ;
	
	// initialize or update based on relative position of cameras and distorted image/camera parameters
	public:
	Rectifier(   const cv::Matx33d rotation,
	             const cv::Vec3d   translation,
	             const Distortion  distortionLeft,
	             const Distortion  distortionRight,
	             const Camera      cameraSrcLeft,
	             const Camera      cameraSrcRight,
	             const cv::Size    sizeSrc,
	             const bool        equalFocalLengths = true ) ;
	void update( const cv::Matx33d rotation,
	             const cv::Vec3d   translation,
	             const Distortion  distortionLeft,
	             const Distortion  distortionRight,
	             const Camera      cameraSrcLeft,
	             const Camera      cameraSrcRight,
	             const cv::Size    sizeSrc,
	             const bool        equalFocalLengths = true ) ;
	
	// rectify: initialize or update based on rectification homographies and distorted/rectified image/camera parameters
	// user note: If "equalFocalLengths" is set, "projDst{Left,Right}Arg" and "sizeDst" will be changed if necessary.
	public:
	Rectifier(   const cv::Matx33d homographyLeft,
	             const cv::Matx33d homographyRight,
	             const Distortion  distortionLeft,
	             const Distortion  distortionRight,
	             const Camera      cameraSrcLeft,
	             const Camera      cameraSrcRight,
	             const Projection  projDstLeftArg,
	             const Projection  projDstRightArg,
	             const cv::Size    sizeDst,
	             const bool        equalFocalLengths = true ) ;
	void update( const cv::Matx33d homographyLeft,
	             const cv::Matx33d homographyRight,
	             const Distortion  distortionLeft,
	             const Distortion  distortionRight,
	             const Camera      cameraSrcLeft,
	             const Camera      cameraSrcRight,
	             const Projection  projDstLeftArg,
	             const Projection  projDstRightArg,
	             const cv::Size    sizeDst,
	             const bool        equalFocalLengths = true ) ;
	
	// initialize or update based on LUTs and rectified projection matrices
	// user note: If "equalFocalLengths" is set, "projDst{Left,Right}Arg" will be changed if necessary.
	public:
	Rectifier(   const cv::Mat    mapULeftArg,
	             const cv::Mat    mapURightArg,
	             const cv::Mat    mapVLeftArg,
	             const cv::Mat    mapVRightArg,
	             const Projection projDstLeftArg,
	             const Projection projDstRightArg,
	             const bool       equalFocalLengths = true ) ;
	void update( const cv::Mat    mapULeftArg,
	             const cv::Mat    mapURightArg,
	             const cv::Mat    mapVLeftArg,
	             const cv::Mat    mapVRightArg,
	             const Projection projDstLeftArg,
	             const Projection projDstRightArg,
	             const bool       equalFocalLengths = true ) ;
	
	// perform transformation of both images on either CPU or GPU
	void CPU( const cv::Mat  srcLeft,
	          const cv::Mat  srcRight,
	                cv::Mat& dstLeft,
	                cv::Mat& dstRight,
	                cv::Rect roi      = cv::Rect(),
	          const int      cores    = hawaii::CPUCores ) const ;
	void GPU( const cv::gpu::GpuMat  srcLeft,
	          const cv::gpu::GpuMat  srcRight,
	                cv::gpu::GpuMat& dstLeft,
	                cv::gpu::GpuMat& dstRight,
	                cv::Rect         roi      = cv::Rect(),
	                cv::gpu::Stream& stream   = cv::gpu::Stream::Null() ) const ;
	
	// get initialization state, output size and projection matrices
	public:
	bool       initialized()        const ;
	cv::Size   sizeDst()            const ;
	Projection projectionDstLeft()  const ;
	Projection projectionDstRight() const ;
	protected:
	Projection projectionDstLM,
	           projectionDstRM ;
	
	// wrapped remappers to actually do the processing, shared mutex for same reason as in "Remapper"
	protected:
	Remapper remapperL,
	         remapperR ;
	mutable boost::shared_mutex mutex ;
} ;

} } // namespaces "toast2::stereo"
