// Copyright (C) 2014 by Karlsruher Institut fuer Technologie (KIT), Institut fuer Mess- und Regelungstechnik
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of arDroneManual/Autonomous.
// 
// arDroneManual/Autonomous is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// arDroneManual/Autonomous is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even 
// the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
// 
// You should have received a copy of the GNU General Public License along with arDroneManual/Autonomous. If not, see 
// <http://www.gnu.org/licenses/>.


// helpers for processing keyboard state
// =====================================

#pragma once

#include "hawaii/common/error.h"

// create a "continuous" value from performing pulse width modulation on the keyboard
class KeySmoother {
	
	// internal state, initialization
	protected:
	float state, step ;
	public:
	KeySmoother( float stepArg = 0.05f ) :
		state( 0.0f    ),
		step(  stepArg ) {
		HAWAII_ERROR_CONDITIONAL( this->step <= 0.0f
		                       || this->step >  1.0f,
		                          "Step must be in (0.0f 1.0f]." ) ;
	}
	
	// update and read: Only pass the state of one key for a result in [0.0f 1.0f]. Use a "more" and a "less" key for a 
	//                  result in [-1.0f 1.0f].
	public:
	float operator ()( const bool more,
	                   const bool less = false ) {
		
		// if one of the keys is pressed, ramp state up or down (but start from zero so that pressing e.g. "forward" 
		// does not need to ramp down an existing "backward" value first) 
		if(  more && !less ) { this->state += this->step ; this->state = std::min(  1.0f, std::max( 0.0f, this->state ) ) ; }
		if( !more &&  less ) { this->state -= this->step ; this->state = std::max( -1.0f, std::min( 0.0f, this->state ) ) ; }
		
		// fade state down to zero if no key is pressed, but implicitly keep it fixed if both are
		if( !more && !less ) {
			if( this->state > 0.0f ) { this->state = std::max( 0.0f, this->state - this->step ) ; }
			else                     { this->state = std::min( 0.0f, this->state + this->step ) ; }
		}
		
		// return the updated value
		return this->state ;
	}
	
	// read without update
	public:
	operator float() { return this->state ; }
	
} ; // class "KeySmoother"
