// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// information on this system's CPUs and GPUs
// ==========================================

#include "hawaii/common/hardware.h"
#include "hawaii/GPU/CUDA.h"
#include "hawaii/common/minMax.h"
#include "hawaii/common/error.h"
#include <opencv2/core/core.hpp>
#include <cstdio>
#include <cstdlib>
#include <map>

// helper(s) only used here
namespace {
	
	// "std::string" obviously isn't good enough for OpenCV 3...
	#if CV_MAJOR_VERSION == 3
		typedef cv::String  String ;
	#elif CV_MAJOR_VERSION == 2
		typedef std::string String ;
	#else
		#error "unsupported version of OpenCV"
	#endif
	
	// return the output of a console command as an integer or a string
	int consoleToInt( const char* command ) {
		char buf[ 4096 ] ;
		FILE* pipe = popen( command, "r" ) ;
		if( fgets( buf, 4096, pipe ) == 0 ) { HAWAII_ERROR( "failed to read from pipe" ) ; }
		pclose( pipe ) ;
		return atoi( buf ) ;
	}
	String consoleToString( const char* command ) {
		char buf[ 4096 ] ;
		FILE* pipe = popen( command, "r" ) ;
		if( fgets( buf, 4096, pipe ) == 0 ) { HAWAII_ERROR( "failed to read from pipe" ) ; }
		pclose( pipe ) ;
		return String( buf ) ;
	}
	
	// map processor name to TDP using a user-local configuration file 
	double processorNameToTDP( const String procName ) {
		
		// name and handle of config file
		const String fileName = String( getenv( "HOME" ) ) + String( "/.libHawaii.yml" ) ;
		cv::FileStorage fileHandle ;
		
		// fill look-up table by reading config file
		// developer note: "std::map" not supported by "cv::FileStorage", so take a detour via "std::vector"
		std::map< String, double > mapProcNameToTDP ;
		if( fileHandle.open( fileName, cv::FileStorage::READ ) ) {
			std::vector< String > names ; fileHandle[ "names" ] >> names ;
			std::vector< double > TDPs  ; fileHandle[ "TDPs"  ] >> TDPs  ;
			for( size_t idx = 0 ; idx < names.size() ; ++idx ) {
				mapProcNameToTDP[ names[ idx ] ] = TDPs[ idx ] ;
			}
			fileHandle.release() ;
		}
		
		// return given processor's TDP if available
		const auto it = mapProcNameToTDP.find( procName ) ;
		if( it != mapProcNameToTDP.end() ) { return it->second ; }
		
		// otherwise...
		else {
			
			// prompt the user for it
			double procPower = NAN ;
			std::cout << "please enter the thermal design power (TDP) of your \"" << procName << "\" processor: " ;
			std::cin >> procPower ;
			
			// update the config file
			mapProcNameToTDP[ procName ] = procPower ;
			fileHandle.open( fileName, cv::FileStorage::WRITE ) ;
			std::vector< String > names ;
			std::vector< double > TDPs  ;
			for( const auto mapEntry : mapProcNameToTDP ) {
				names.push_back( mapEntry.first  ) ;
				TDPs.push_back(  mapEntry.second ) ;
			}
			fileHandle << "names" << names ;
			fileHandle << "TDPs"  << TDPs  ;
			fileHandle.release() ;
			
			// return the entered TDP
			return procPower ;
		}
		
	} // function "processorNameToTDP()"
	
	// number of physical CPUs, real cores and register sets, thermal design power of all CPUs combined
	// developer note: The number of cores per socket cannot be determined by a single command because of 2 cases:
	//                 a) For multiple silicon dice within one case, evaluating by "core id" only counts one die.
	//                 b) For "modules" with multiple processing units, evaluating by "cpu cores" only counts one unit.
	int CPUSocketsImpl()          { return           consoleToInt( "cat /proc/cpuinfo | grep \"physical id\" | sort -u | wc -l"            )   ; }
	int CPUCoresPerSocketImpl()   { return std::max( consoleToInt( "cat /proc/cpuinfo | grep \"core id\"     | sort -u | wc -l"            ),
	                                                 consoleToInt( "cat /proc/cpuinfo | grep \"cpu cores\"   | sort -u | awk '{print $4}'" ) ) ; }
	int CPUThreadsPerSocketImpl() { return           consoleToInt( "cat /proc/cpuinfo | grep \"siblings\"    | sort -u | awk '{print $3}'" )   ; }
	double CPUPowerPerSocketImpl() {
		String CPUName = consoleToString( "cat /proc/cpuinfo | grep \"model name\" | sort -u | awk '{ for( i = 4 ; i < NF ; ++i ) printf( \"%s \",$i ) ; printf( \"%s\", $NF ) }'" ) ;
		return processorNameToTDP( CPUName ) ;
	}
	
	// number of GPUs present in the system
	// user note: 0 if compile-time GPU support is disabled, also limited by "HAWAII_GPUS_MAX"
	int GPUsImpl() {
		#ifndef HAWAII_NO_CUDA
			int GPUs ; if( cudaSuccess != cudaGetDeviceCount( &GPUs ) ) { return 0 ; }
			if( GPUs > HAWAII_GPUS_MAX ) { GPUs = HAWAII_GPUS_MAX ; }
			return GPUs ;
		#else
			return 0 ;
		#endif
	}
	
	// determine if GPU has own instead of sharing CPU memory
	bool GPUMemoryOwnImpl( int GPU ) {
		#ifndef HAWAII_NO_CUDA
			if( HAWAII_UNLIKELY( (uint)GPU >= (uint)GPUsImpl() ) ) { return false ; }
			int integrated       ; if( cudaSuccess != cudaDeviceGetAttribute( &integrated,       cudaDevAttrIntegrated,       GPU ) ) { return false ; }
			int canMapHostMemory ; if( cudaSuccess != cudaDeviceGetAttribute( &canMapHostMemory, cudaDevAttrCanMapHostMemory, GPU ) ) { return false ; }
			return !( integrated && canMapHostMemory ) ;
		#else
			(void)GPU ;
			return false ;
		#endif
	}
	
	// GPU's compute capability: same format as NVCC's "__CUDA_ARCH__" macro, i.e. CC 2.1 yields 210
	int GPUComputeCapabilityImpl( int GPU ) {
		#ifndef HAWAII_NO_CUDA
			if( HAWAII_UNLIKELY( (uint)GPU >= (uint)GPUsImpl() ) ) { return -1 ; }
			int computeCapabilityMajor ; if( cudaSuccess != cudaDeviceGetAttribute( &computeCapabilityMajor, cudaDevAttrComputeCapabilityMajor, GPU ) ) { return -1 ; }
			int computeCapabilityMinor ; if( cudaSuccess != cudaDeviceGetAttribute( &computeCapabilityMinor, cudaDevAttrComputeCapabilityMinor, GPU ) ) { return -1 ; }
			return ( 100 * computeCapabilityMajor
			       +  10 * computeCapabilityMinor ) ;
		#else
			(void)GPU ;
			return -1 ;
		#endif
	}
	
	// GPU's SIMD support: true if native hardware support for processing two 16-bit or four 8-bit integers within one
	//                     register is available (speed-up of 1.7 to 4.0), false if emulated (speed-"up" of ~0.2)
	// developer note: return false only if you find the phrase "multiple instructions" at "SIMD video instructions" on 
	//                 "http://docs.nvidia.com/cuda/cuda-c-programming-guide/index.html#maximize-instruction-
	//                 throughput".
	bool GPUSIMDSupportImpl( int GPU ) {
		#ifndef HAWAII_NO_CUDA
			if( HAWAII_UNLIKELY( (uint)GPU >= (uint)GPUsImpl() ) ) { return -1 ; }
			const int computeCapability = GPUComputeCapabilityImpl( GPU ) ;
			return ( computeCapability != 500
			      && computeCapability != 520 ) ;
		#else
			(void)GPU ;
			return true ;
		#endif
	}
	
	// minimum required alignment for textures or surfaces
	size_t GPUAlignmentImpl( int GPU ) {
		#ifndef HAWAII_NO_CUDA
			if( HAWAII_UNLIKELY( (uint)GPU >= (uint)GPUsImpl() ) ) { return 1 ; }
			cudaDeviceProp prop ; if( cudaSuccess != cudaGetDeviceProperties( &prop, GPU ) ) { return 1 ; }
			return hawaii::max( prop.textureAlignment, prop.texturePitchAlignment, prop.surfaceAlignment ) ;
		#else
			(void)GPU ;
			return 1 ;
		#endif
	}
	
	// GPU's support for unified addressing
	bool GPUUnifiedAddressingImpl( int GPU ) {
		#ifndef HAWAII_NO_CUDA
			if( HAWAII_UNLIKELY( (uint)GPU >= (uint)GPUsImpl() ) ) { return false ; }
			int unifiedAddressing ; if( cudaSuccess != cudaDeviceGetAttribute( &unifiedAddressing, cudaDevAttrUnifiedAddressing, GPU ) ) { return false ; }
			return ( unifiedAddressing > 0 ) ;
		#else
			(void)GPU ;
			return false ;
		#endif
	}
	
	// thermal design power of each GPU
	double GPUPowerImpl( int GPU ) {
		#ifndef HAWAII_NO_CUDA
			if( HAWAII_UNLIKELY( (uint)GPU >= (uint)GPUsImpl() ) ) { return NAN ; }
			cudaDeviceProp prop ;
			cudaGetDeviceProperties( &prop, GPU ) ;
			return processorNameToTDP( prop.name ) ;
		#else
			(void)GPU ;
			return NAN ;
		#endif
	}
	
	// cache results from the functions above
	// developer note: Don't rely on initialization order!
	const bool   GPUMemoryOwnArray[         HAWAII_GPUS_MAX ] = { GPUMemoryOwnImpl(          0 ), GPUMemoryOwnImpl(          1 ), GPUMemoryOwnImpl(          2 ), GPUMemoryOwnImpl(          3 ),
	                                                              GPUMemoryOwnImpl(          4 ), GPUMemoryOwnImpl(          5 ), GPUMemoryOwnImpl(          6 ), GPUMemoryOwnImpl(          7 ),
	                                                              GPUMemoryOwnImpl(          8 ), GPUMemoryOwnImpl(          9 ), GPUMemoryOwnImpl(         10 ), GPUMemoryOwnImpl(         11 ),
	                                                              GPUMemoryOwnImpl(         12 ), GPUMemoryOwnImpl(         13 ), GPUMemoryOwnImpl(         14 ), GPUMemoryOwnImpl(         15 ),
	                                                              GPUMemoryOwnImpl(         16 ), GPUMemoryOwnImpl(         17 ), GPUMemoryOwnImpl(         18 ), GPUMemoryOwnImpl(         19 ),
	                                                              GPUMemoryOwnImpl(         20 ), GPUMemoryOwnImpl(         21 ), GPUMemoryOwnImpl(         22 ), GPUMemoryOwnImpl(         23 ),
	                                                              GPUMemoryOwnImpl(         24 ), GPUMemoryOwnImpl(         25 ), GPUMemoryOwnImpl(         26 ), GPUMemoryOwnImpl(         27 ),
	                                                              GPUMemoryOwnImpl(         28 ), GPUMemoryOwnImpl(         29 ), GPUMemoryOwnImpl(         30 ), GPUMemoryOwnImpl(         31 ),
	                                                              GPUMemoryOwnImpl(         32 ), GPUMemoryOwnImpl(         33 ), GPUMemoryOwnImpl(         34 ), GPUMemoryOwnImpl(         35 ),
	                                                              GPUMemoryOwnImpl(         36 ), GPUMemoryOwnImpl(         37 ), GPUMemoryOwnImpl(         38 ), GPUMemoryOwnImpl(         39 ),
	                                                              GPUMemoryOwnImpl(         40 ), GPUMemoryOwnImpl(         41 ), GPUMemoryOwnImpl(         42 ), GPUMemoryOwnImpl(         43 ),
	                                                              GPUMemoryOwnImpl(         44 ), GPUMemoryOwnImpl(         45 ), GPUMemoryOwnImpl(         46 ), GPUMemoryOwnImpl(         47 ),
	                                                              GPUMemoryOwnImpl(         48 ), GPUMemoryOwnImpl(         49 ), GPUMemoryOwnImpl(         50 ), GPUMemoryOwnImpl(         51 ),
	                                                              GPUMemoryOwnImpl(         52 ), GPUMemoryOwnImpl(         53 ), GPUMemoryOwnImpl(         54 ), GPUMemoryOwnImpl(         55 ),
	                                                              GPUMemoryOwnImpl(         56 ), GPUMemoryOwnImpl(         57 ), GPUMemoryOwnImpl(         58 ), GPUMemoryOwnImpl(         59 ),
	                                                              GPUMemoryOwnImpl(         60 ), GPUMemoryOwnImpl(         61 ), GPUMemoryOwnImpl(         62 ), GPUMemoryOwnImpl(         63 ) } ;
	const int    GPUComputeCapabilityArray[ HAWAII_GPUS_MAX ] = { GPUComputeCapabilityImpl(  0 ), GPUComputeCapabilityImpl(  1 ), GPUComputeCapabilityImpl(  2 ), GPUComputeCapabilityImpl(  3 ),
	                                                              GPUComputeCapabilityImpl(  4 ), GPUComputeCapabilityImpl(  5 ), GPUComputeCapabilityImpl(  6 ), GPUComputeCapabilityImpl(  7 ),
	                                                              GPUComputeCapabilityImpl(  8 ), GPUComputeCapabilityImpl(  9 ), GPUComputeCapabilityImpl( 10 ), GPUComputeCapabilityImpl( 11 ),
	                                                              GPUComputeCapabilityImpl( 12 ), GPUComputeCapabilityImpl( 13 ), GPUComputeCapabilityImpl( 14 ), GPUComputeCapabilityImpl( 15 ),
	                                                              GPUComputeCapabilityImpl( 16 ), GPUComputeCapabilityImpl( 17 ), GPUComputeCapabilityImpl( 18 ), GPUComputeCapabilityImpl( 19 ),
	                                                              GPUComputeCapabilityImpl( 20 ), GPUComputeCapabilityImpl( 21 ), GPUComputeCapabilityImpl( 22 ), GPUComputeCapabilityImpl( 23 ),
	                                                              GPUComputeCapabilityImpl( 24 ), GPUComputeCapabilityImpl( 25 ), GPUComputeCapabilityImpl( 26 ), GPUComputeCapabilityImpl( 27 ),
	                                                              GPUComputeCapabilityImpl( 28 ), GPUComputeCapabilityImpl( 29 ), GPUComputeCapabilityImpl( 30 ), GPUComputeCapabilityImpl( 31 ),
	                                                              GPUComputeCapabilityImpl( 32 ), GPUComputeCapabilityImpl( 33 ), GPUComputeCapabilityImpl( 34 ), GPUComputeCapabilityImpl( 35 ),
	                                                              GPUComputeCapabilityImpl( 36 ), GPUComputeCapabilityImpl( 37 ), GPUComputeCapabilityImpl( 38 ), GPUComputeCapabilityImpl( 39 ),
	                                                              GPUComputeCapabilityImpl( 40 ), GPUComputeCapabilityImpl( 41 ), GPUComputeCapabilityImpl( 42 ), GPUComputeCapabilityImpl( 43 ),
	                                                              GPUComputeCapabilityImpl( 44 ), GPUComputeCapabilityImpl( 45 ), GPUComputeCapabilityImpl( 46 ), GPUComputeCapabilityImpl( 47 ),
	                                                              GPUComputeCapabilityImpl( 48 ), GPUComputeCapabilityImpl( 49 ), GPUComputeCapabilityImpl( 50 ), GPUComputeCapabilityImpl( 51 ),
	                                                              GPUComputeCapabilityImpl( 52 ), GPUComputeCapabilityImpl( 53 ), GPUComputeCapabilityImpl( 54 ), GPUComputeCapabilityImpl( 55 ),
	                                                              GPUComputeCapabilityImpl( 56 ), GPUComputeCapabilityImpl( 57 ), GPUComputeCapabilityImpl( 58 ), GPUComputeCapabilityImpl( 59 ),
	                                                              GPUComputeCapabilityImpl( 60 ), GPUComputeCapabilityImpl( 61 ), GPUComputeCapabilityImpl( 62 ), GPUComputeCapabilityImpl( 63 ) } ;
	const bool   GPUSIMDSupportArray[       HAWAII_GPUS_MAX ] = { GPUSIMDSupportImpl(        0 ), GPUSIMDSupportImpl(        1 ), GPUSIMDSupportImpl(        2 ), GPUSIMDSupportImpl(        3 ),
	                                                              GPUSIMDSupportImpl(        4 ), GPUSIMDSupportImpl(        5 ), GPUSIMDSupportImpl(        6 ), GPUSIMDSupportImpl(        7 ),
	                                                              GPUSIMDSupportImpl(        8 ), GPUSIMDSupportImpl(        9 ), GPUSIMDSupportImpl(       10 ), GPUSIMDSupportImpl(       11 ),
	                                                              GPUSIMDSupportImpl(       12 ), GPUSIMDSupportImpl(       13 ), GPUSIMDSupportImpl(       14 ), GPUSIMDSupportImpl(       15 ),
	                                                              GPUSIMDSupportImpl(       16 ), GPUSIMDSupportImpl(       17 ), GPUSIMDSupportImpl(       18 ), GPUSIMDSupportImpl(       19 ),
	                                                              GPUSIMDSupportImpl(       20 ), GPUSIMDSupportImpl(       21 ), GPUSIMDSupportImpl(       22 ), GPUSIMDSupportImpl(       23 ),
	                                                              GPUSIMDSupportImpl(       24 ), GPUSIMDSupportImpl(       25 ), GPUSIMDSupportImpl(       26 ), GPUSIMDSupportImpl(       27 ),
	                                                              GPUSIMDSupportImpl(       28 ), GPUSIMDSupportImpl(       29 ), GPUSIMDSupportImpl(       30 ), GPUSIMDSupportImpl(       31 ),
	                                                              GPUSIMDSupportImpl(       32 ), GPUSIMDSupportImpl(       33 ), GPUSIMDSupportImpl(       34 ), GPUSIMDSupportImpl(       35 ),
	                                                              GPUSIMDSupportImpl(       36 ), GPUSIMDSupportImpl(       37 ), GPUSIMDSupportImpl(       38 ), GPUSIMDSupportImpl(       39 ),
	                                                              GPUSIMDSupportImpl(       40 ), GPUSIMDSupportImpl(       41 ), GPUSIMDSupportImpl(       42 ), GPUSIMDSupportImpl(       43 ),
	                                                              GPUSIMDSupportImpl(       44 ), GPUSIMDSupportImpl(       45 ), GPUSIMDSupportImpl(       46 ), GPUSIMDSupportImpl(       47 ),
	                                                              GPUSIMDSupportImpl(       48 ), GPUSIMDSupportImpl(       49 ), GPUSIMDSupportImpl(       50 ), GPUSIMDSupportImpl(       51 ),
	                                                              GPUSIMDSupportImpl(       52 ), GPUSIMDSupportImpl(       53 ), GPUSIMDSupportImpl(       54 ), GPUSIMDSupportImpl(       55 ),
	                                                              GPUSIMDSupportImpl(       56 ), GPUSIMDSupportImpl(       57 ), GPUSIMDSupportImpl(       58 ), GPUSIMDSupportImpl(       59 ),
	                                                              GPUSIMDSupportImpl(       60 ), GPUSIMDSupportImpl(       61 ), GPUSIMDSupportImpl(       62 ), GPUSIMDSupportImpl(       63 ) } ;
	const size_t GPUAlignmentArray[         HAWAII_GPUS_MAX ] = { GPUAlignmentImpl(          0 ), GPUAlignmentImpl(          1 ), GPUAlignmentImpl(          2 ), GPUAlignmentImpl(          3 ),
	                                                              GPUAlignmentImpl(          4 ), GPUAlignmentImpl(          5 ), GPUAlignmentImpl(          6 ), GPUAlignmentImpl(          7 ),
	                                                              GPUAlignmentImpl(          8 ), GPUAlignmentImpl(          9 ), GPUAlignmentImpl(         10 ), GPUAlignmentImpl(         11 ),
	                                                              GPUAlignmentImpl(         12 ), GPUAlignmentImpl(         13 ), GPUAlignmentImpl(         14 ), GPUAlignmentImpl(         15 ),
	                                                              GPUAlignmentImpl(         16 ), GPUAlignmentImpl(         17 ), GPUAlignmentImpl(         18 ), GPUAlignmentImpl(         19 ),
	                                                              GPUAlignmentImpl(         20 ), GPUAlignmentImpl(         21 ), GPUAlignmentImpl(         22 ), GPUAlignmentImpl(         23 ),
	                                                              GPUAlignmentImpl(         24 ), GPUAlignmentImpl(         25 ), GPUAlignmentImpl(         26 ), GPUAlignmentImpl(         27 ),
	                                                              GPUAlignmentImpl(         28 ), GPUAlignmentImpl(         29 ), GPUAlignmentImpl(         30 ), GPUAlignmentImpl(         31 ),
	                                                              GPUAlignmentImpl(         32 ), GPUAlignmentImpl(         33 ), GPUAlignmentImpl(         34 ), GPUAlignmentImpl(         35 ),
	                                                              GPUAlignmentImpl(         36 ), GPUAlignmentImpl(         37 ), GPUAlignmentImpl(         38 ), GPUAlignmentImpl(         39 ),
	                                                              GPUAlignmentImpl(         40 ), GPUAlignmentImpl(         41 ), GPUAlignmentImpl(         42 ), GPUAlignmentImpl(         43 ),
	                                                              GPUAlignmentImpl(         44 ), GPUAlignmentImpl(         45 ), GPUAlignmentImpl(         46 ), GPUAlignmentImpl(         47 ),
	                                                              GPUAlignmentImpl(         48 ), GPUAlignmentImpl(         49 ), GPUAlignmentImpl(         50 ), GPUAlignmentImpl(         51 ),
	                                                              GPUAlignmentImpl(         52 ), GPUAlignmentImpl(         53 ), GPUAlignmentImpl(         54 ), GPUAlignmentImpl(         55 ),
	                                                              GPUAlignmentImpl(         56 ), GPUAlignmentImpl(         57 ), GPUAlignmentImpl(         58 ), GPUAlignmentImpl(         59 ),
	                                                              GPUAlignmentImpl(         60 ), GPUAlignmentImpl(         61 ), GPUAlignmentImpl(         62 ), GPUAlignmentImpl(         63 ) } ;
	const size_t GPUAlignmentMaxInstance = hawaii::max(           GPUAlignmentImpl(          0 ), GPUAlignmentImpl(          1 ), GPUAlignmentImpl(          2 ), GPUAlignmentImpl(          3 ),
	                                                              GPUAlignmentImpl(          4 ), GPUAlignmentImpl(          5 ), GPUAlignmentImpl(          6 ), GPUAlignmentImpl(          7 ),
	                                                              GPUAlignmentImpl(          8 ), GPUAlignmentImpl(          9 ), GPUAlignmentImpl(         10 ), GPUAlignmentImpl(         11 ),
	                                                              GPUAlignmentImpl(         12 ), GPUAlignmentImpl(         13 ), GPUAlignmentImpl(         14 ), GPUAlignmentImpl(         15 ),
	                                                              GPUAlignmentImpl(         16 ), GPUAlignmentImpl(         17 ), GPUAlignmentImpl(         18 ), GPUAlignmentImpl(         19 ),
	                                                              GPUAlignmentImpl(         20 ), GPUAlignmentImpl(         21 ), GPUAlignmentImpl(         22 ), GPUAlignmentImpl(         23 ),
	                                                              GPUAlignmentImpl(         24 ), GPUAlignmentImpl(         25 ), GPUAlignmentImpl(         26 ), GPUAlignmentImpl(         27 ),
	                                                              GPUAlignmentImpl(         28 ), GPUAlignmentImpl(         29 ), GPUAlignmentImpl(         30 ), GPUAlignmentImpl(         31 ),
	                                                              GPUAlignmentImpl(         32 ), GPUAlignmentImpl(         33 ), GPUAlignmentImpl(         34 ), GPUAlignmentImpl(         35 ),
	                                                              GPUAlignmentImpl(         36 ), GPUAlignmentImpl(         37 ), GPUAlignmentImpl(         38 ), GPUAlignmentImpl(         39 ),
	                                                              GPUAlignmentImpl(         40 ), GPUAlignmentImpl(         41 ), GPUAlignmentImpl(         42 ), GPUAlignmentImpl(         43 ),
	                                                              GPUAlignmentImpl(         44 ), GPUAlignmentImpl(         45 ), GPUAlignmentImpl(         46 ), GPUAlignmentImpl(         47 ),
	                                                              GPUAlignmentImpl(         48 ), GPUAlignmentImpl(         49 ), GPUAlignmentImpl(         50 ), GPUAlignmentImpl(         51 ),
	                                                              GPUAlignmentImpl(         52 ), GPUAlignmentImpl(         53 ), GPUAlignmentImpl(         54 ), GPUAlignmentImpl(         55 ),
	                                                              GPUAlignmentImpl(         56 ), GPUAlignmentImpl(         57 ), GPUAlignmentImpl(         58 ), GPUAlignmentImpl(         59 ),
	                                                              GPUAlignmentImpl(         60 ), GPUAlignmentImpl(         61 ), GPUAlignmentImpl(         62 ), GPUAlignmentImpl(         63 ) ) ;
	const bool   GPUUnifiedAddressingArray[ HAWAII_GPUS_MAX ] = { GPUUnifiedAddressingImpl(  0 ), GPUUnifiedAddressingImpl(  1 ), GPUUnifiedAddressingImpl(  2 ), GPUUnifiedAddressingImpl(  3 ),
	                                                              GPUUnifiedAddressingImpl(  4 ), GPUUnifiedAddressingImpl(  5 ), GPUUnifiedAddressingImpl(  6 ), GPUUnifiedAddressingImpl(  7 ),
	                                                              GPUUnifiedAddressingImpl(  8 ), GPUUnifiedAddressingImpl(  9 ), GPUUnifiedAddressingImpl( 10 ), GPUUnifiedAddressingImpl( 11 ),
	                                                              GPUUnifiedAddressingImpl( 12 ), GPUUnifiedAddressingImpl( 13 ), GPUUnifiedAddressingImpl( 14 ), GPUUnifiedAddressingImpl( 15 ),
	                                                              GPUUnifiedAddressingImpl( 16 ), GPUUnifiedAddressingImpl( 17 ), GPUUnifiedAddressingImpl( 18 ), GPUUnifiedAddressingImpl( 19 ),
	                                                              GPUUnifiedAddressingImpl( 20 ), GPUUnifiedAddressingImpl( 21 ), GPUUnifiedAddressingImpl( 22 ), GPUUnifiedAddressingImpl( 23 ),
	                                                              GPUUnifiedAddressingImpl( 24 ), GPUUnifiedAddressingImpl( 25 ), GPUUnifiedAddressingImpl( 26 ), GPUUnifiedAddressingImpl( 27 ),
	                                                              GPUUnifiedAddressingImpl( 28 ), GPUUnifiedAddressingImpl( 29 ), GPUUnifiedAddressingImpl( 30 ), GPUUnifiedAddressingImpl( 31 ),
	                                                              GPUUnifiedAddressingImpl( 32 ), GPUUnifiedAddressingImpl( 33 ), GPUUnifiedAddressingImpl( 34 ), GPUUnifiedAddressingImpl( 35 ),
	                                                              GPUUnifiedAddressingImpl( 36 ), GPUUnifiedAddressingImpl( 37 ), GPUUnifiedAddressingImpl( 38 ), GPUUnifiedAddressingImpl( 39 ),
	                                                              GPUUnifiedAddressingImpl( 40 ), GPUUnifiedAddressingImpl( 41 ), GPUUnifiedAddressingImpl( 42 ), GPUUnifiedAddressingImpl( 43 ),
	                                                              GPUUnifiedAddressingImpl( 44 ), GPUUnifiedAddressingImpl( 45 ), GPUUnifiedAddressingImpl( 46 ), GPUUnifiedAddressingImpl( 47 ),
	                                                              GPUUnifiedAddressingImpl( 48 ), GPUUnifiedAddressingImpl( 49 ), GPUUnifiedAddressingImpl( 50 ), GPUUnifiedAddressingImpl( 51 ),
	                                                              GPUUnifiedAddressingImpl( 52 ), GPUUnifiedAddressingImpl( 53 ), GPUUnifiedAddressingImpl( 54 ), GPUUnifiedAddressingImpl( 55 ),
	                                                              GPUUnifiedAddressingImpl( 56 ), GPUUnifiedAddressingImpl( 57 ), GPUUnifiedAddressingImpl( 58 ), GPUUnifiedAddressingImpl( 59 ),
	                                                              GPUUnifiedAddressingImpl( 60 ), GPUUnifiedAddressingImpl( 61 ), GPUUnifiedAddressingImpl( 62 ), GPUUnifiedAddressingImpl( 63 ) } ;
	const double GPUPowerArray[             HAWAII_GPUS_MAX ] = { GPUPowerImpl(              0 ), GPUPowerImpl(              1 ), GPUPowerImpl(              2 ), GPUPowerImpl(              3 ),
	                                                              GPUPowerImpl(              4 ), GPUPowerImpl(              5 ), GPUPowerImpl(              6 ), GPUPowerImpl(              7 ),
	                                                              GPUPowerImpl(              8 ), GPUPowerImpl(              9 ), GPUPowerImpl(             10 ), GPUPowerImpl(             11 ),
	                                                              GPUPowerImpl(             12 ), GPUPowerImpl(             13 ), GPUPowerImpl(             14 ), GPUPowerImpl(             15 ),
	                                                              GPUPowerImpl(             16 ), GPUPowerImpl(             17 ), GPUPowerImpl(             18 ), GPUPowerImpl(             19 ),
	                                                              GPUPowerImpl(             20 ), GPUPowerImpl(             21 ), GPUPowerImpl(             22 ), GPUPowerImpl(             23 ),
	                                                              GPUPowerImpl(             24 ), GPUPowerImpl(             25 ), GPUPowerImpl(             26 ), GPUPowerImpl(             27 ),
	                                                              GPUPowerImpl(             28 ), GPUPowerImpl(             29 ), GPUPowerImpl(             30 ), GPUPowerImpl(             31 ),
	                                                              GPUPowerImpl(             32 ), GPUPowerImpl(             33 ), GPUPowerImpl(             34 ), GPUPowerImpl(             35 ),
	                                                              GPUPowerImpl(             36 ), GPUPowerImpl(             37 ), GPUPowerImpl(             38 ), GPUPowerImpl(             39 ),
	                                                              GPUPowerImpl(             40 ), GPUPowerImpl(             41 ), GPUPowerImpl(             42 ), GPUPowerImpl(             43 ),
	                                                              GPUPowerImpl(             44 ), GPUPowerImpl(             45 ), GPUPowerImpl(             46 ), GPUPowerImpl(             47 ),
	                                                              GPUPowerImpl(             48 ), GPUPowerImpl(             49 ), GPUPowerImpl(             50 ), GPUPowerImpl(             51 ),
	                                                              GPUPowerImpl(             52 ), GPUPowerImpl(             53 ), GPUPowerImpl(             54 ), GPUPowerImpl(             55 ),
	                                                              GPUPowerImpl(             56 ), GPUPowerImpl(             57 ), GPUPowerImpl(             58 ), GPUPowerImpl(             59 ),
	                                                              GPUPowerImpl(             60 ), GPUPowerImpl(             61 ), GPUPowerImpl(             62 ), GPUPowerImpl(             63 ) } ;
	} // anonymous namespace

namespace hawaii {

// number of physical CPUs, real cores and register sets, thermal design power of all CPUs combined
// developer note: Don't rely on initialization order!
const int    CPUSockets          = CPUSocketsImpl()                             ;
const int    CPUCores            = CPUSocketsImpl() * CPUCoresPerSocketImpl()   ;
const int    CPUCoresPerSocket   =                    CPUCoresPerSocketImpl()   ;
const int    CPUThreads          = CPUSocketsImpl() * CPUThreadsPerSocketImpl() ;
const int    CPUThreadsPerSocket =                    CPUThreadsPerSocketImpl() ;
const double CPUPower            = CPUSocketsImpl() * CPUPowerPerSocketImpl()   ;

// number of GPUs present in the system
extern const int GPUs = GPUsImpl() ;

// check maximum number of GPUs supported by "libHawaii"
static_assert( HAWAII_GPUS_MAX == 64, "Changing \"HAWAII_GPU_GPUSMAX\" requires further modifications." ) ;

// determine if GPU has own instead of sharing CPU memory
bool GPUMemoryOwn( int GPU ) {
	HAWAII_ERROR_NO_CUDA ;
	#ifndef HAWAII_NO_CUDA
		if( GPU < 0 ) { HAWAII_ERROR_CHECK_CUDA(  cudaGetDevice( &GPU )                        ) ; }
		else          { HAWAII_ERROR_CONDITIONAL( (uint)GPU >= (uint)GPUs, "invalid GPU index" ) ; }
	#endif
	return GPUMemoryOwnArray[ GPU ] ;
}

// GPU's compute capability: same format as NVCC's "__CUDA_ARCH__" macro, i.e. CC 2.1 yields 210
int GPUComputeCapability( int GPU ) {
	HAWAII_ERROR_NO_CUDA ;
	#ifndef HAWAII_NO_CUDA
		if( GPU < 0 ) { HAWAII_ERROR_CHECK_CUDA(  cudaGetDevice( &GPU )                        ) ; }
		else          { HAWAII_ERROR_CONDITIONAL( (uint)GPU >= (uint)GPUs, "invalid GPU index" ) ; }
	#endif
	return GPUComputeCapabilityArray[ GPU ] ;
}

// GPU's SIMD support: true if native hardware support for processing two 16-bit or four 8-bit integers within one
//                     register is available (speed-up of 1.7 to 4.0), false if emulated (speed-"up" of ~0.2)
bool GPUSIMDSupport( int GPU ) {
	HAWAII_ERROR_NO_CUDA ;
	#ifndef HAWAII_NO_CUDA
		if( GPU < 0 ) { HAWAII_ERROR_CHECK_CUDA(  cudaGetDevice( &GPU )                        ) ; }
		else          { HAWAII_ERROR_CONDITIONAL( (uint)GPU >= (uint)GPUs, "invalid GPU index" ) ; }
	#endif
	return GPUSIMDSupportArray[ GPU ] ;
}

// minimum required alignment for textures or surfaces, either for a specific GPU or the maximum among all GPUs
size_t GPUAlignment( int GPU ) {
	HAWAII_ERROR_NO_CUDA ;
	#ifndef HAWAII_NO_CUDA
		if( GPU < 0 ) { HAWAII_ERROR_CHECK_CUDA(  cudaGetDevice( &GPU )                        ) ; }
		else          { HAWAII_ERROR_CONDITIONAL( (uint)GPU >= (uint)GPUs, "invalid GPU index" ) ; }
	#endif
	return GPUAlignmentArray[ GPU ] ;
}
size_t GPUAlignmentMax() {
	HAWAII_ERROR_NO_CUDA ;
	return GPUAlignmentMaxInstance ;
}

// GPU's support for unified addressing
bool GPUUnifiedAddressing( int GPU ) {
	HAWAII_ERROR_NO_CUDA ;
	#ifndef HAWAII_NO_CUDA
		if( GPU < 0 ) { HAWAII_ERROR_CHECK_CUDA(  cudaGetDevice( &GPU )                        ) ; }
		else          { HAWAII_ERROR_CONDITIONAL( (uint)GPU >= (uint)GPUs, "invalid GPU index" ) ; }
	#endif
	return GPUUnifiedAddressingArray[ GPU ] ;
}

// thermal design power of each GPU
double GPUPower( int GPU ) {
	HAWAII_ERROR_NO_CUDA ;
	#ifndef HAWAII_NO_CUDA
		if( GPU < 0 ) { HAWAII_ERROR_CHECK_CUDA(  cudaGetDevice( &GPU )                        ) ; }
		else          { HAWAII_ERROR_CONDITIONAL( (uint)GPU >= (uint)GPUs, "invalid GPU index" ) ; }
	#endif
	return GPUPowerArray[ GPU ] ;
}

} // namespace "hawaii"
