// Copyright (C) 2014 by Karlsruher Institut fuer Technologie (KIT), Institut fuer Mess- und Regelungstechnik
// 
// Authors: Benjamin Ranft (benjamin.ranft@web.de), Jacques Maylaender
// 
// This file is part of arDroneManual/Autonomous.
// 
// arDroneManual/Autonomous is free software: you can redistribute it and/or modify it under the terms of the GNU 
// General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your 
// option) any later version.
// 
// arDroneManual/Autonomous is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even 
// the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
// more details.
// 
// You should have received a copy of the GNU General Public License along with arDroneManual/Autonomous. If not, see 
// <http://www.gnu.org/licenses/>.


// line detection
// ==============

// note: "arDroneAutonomous/src/lineDetection.cpp" and "offLineDetection/src/lineDetection.cpp" are symlinked, i.e. 
//       modifications occur in both at the same time.

#include "lineDetection.h"
#include "clozer.h"
#include "hawaii/common/timer.h"
#include "hawaii/common/helpers.h"
#include "hawaii/common/ostream.h"
#include "hawaii/common/error.h"
#include <opencv2/imgproc/imgproc.hpp>
#include <iostream>
#include <cmath>

// helpers to handle a pixel's color channels
struct BGR8 { uint8_t b, g, r ; } ;
struct HSV8 { uint8_t h, s, v ; } ;

// helper for image point coordinate: column "u" and row "v"
struct Point {
	int u, v ;
	Point( const int uArg,
	       const int vArg ) :
		u( uArg ),
		v( vArg ) {
	}
} ;

// helper to draw line points in orange
void drawLinePoints( const std::vector< Point >& linePoints,
                           cv::Mat&              visu       ) ;

// helper to draw line points in orange, with opacity proportional to weight
void drawWeightedLinePoints( const std::vector< Point >& linePoints,
                             const cv::Mat_< double >    weights,
                                   cv::Mat&              visu       ) ;

// helper to draw RANSAC seed points as big yellow X and supporters in orange
void drawRANSACLinePoints( const std::vector< Point >&  linePoints,
                           const std::vector< size_t >& seeds,
                           const std::vector< size_t >& supporters,
                                 cv::Mat&               visu       ) ;

// helper to draw a polynomial in yellow
void drawLineModel( const cv::Mat_< double > coeffs,
                          cv::Mat&           visu   ) ;


// estimate 3rd order polynomial coefficients from an image of a roughly vertical red line
Coeffs detectLine( const cv::Mat image ) {
	
	// check inputs
	if( image.empty() ) {
		hawaii::cerr << "ERROR: The input image must not be empty." << hawaii::endl ;
		return Coeffs() ;
	}
	if( image.type() != CV_8UC3 ) {
		hawaii::cerr << "ERROR: The input image must contain 8-bit unsigned three-channel pixels." << hawaii::endl ;
		return Coeffs() ;
	}
	
	// create a copy of the input image to draw visualizations on
	cv::Mat visu ;
	image.copyTo( visu ) ;
	
	// only in debug mode, show input image and wait for a keystroke
	#ifdef DEBUG
		cv::imshow( "line detection", visu ) ;
		cv::waitKey() ;
	#endif
	
	// 1) compute lineness image ("Je heller, desto Linie")
	// ====================================================
	
	// convert input image from BGR (blue, green, red) to HSV (hue, saturation, value) color space
	// hints: - The hue channel has values in [0, 180) to cover the 360° color circle. Blue is ca. 120 (i.e. 
	//          240°), green is ca. 60 (i.e. 120°). The wrap-around from 360° back to 0° occurs at red, so its values 
	//          are either 0 + X or 180 - X. 
	//        - Saturation goes from 0 for gray to 255 for vivid color.
	//        - Value goes from 0 for black to 255 for bright color.
	cv::Mat hsv ;
	cv::cvtColor( image, hsv, CV_BGR2HSV ) ; // STUD_FUNCTION
	
	// create lineness matrix with the same size as the HSV image
	cv::Mat lineness( hsv.size(), CV_8UC1 ) ;
	
	// two lineness computation modes available
	enum {
		linenessModeThreshold,
		linenessModeFuzzy
	} linenessMode = linenessModeThreshold ; // STUD_TODO try both modes
	switch( linenessMode ) {
		
		// mode A: binary thresholds
		case linenessModeThreshold : {
			
			// find red-orange, i.e. hue in [0°, 30°) and high enough saturation and value
			cv::Mat orange ;
			cv::inRange( hsv,
			             cv::Scalar( STUD_NAN(  0/2 ),  75, 100 ), // STUD_TODO define minimum...
			             cv::Scalar( STUD_NAN( 30/2 ), 255, 255 ), // ...and maximum hue
			             orange                        ) ;
			
			// find red-violet, i.e. hue in [330°, 360°) and high enough saturation and value
			cv::Mat violet ;
			cv::inRange( hsv,
			             cv::Scalar( 330/2, STUD_NAN(  75 ), 100 ), // STUD_TODO define minimum...
			             cv::Scalar( 360/2, STUD_NAN( 255 ), 255 ), // ...and maximum saturation
			             violet                        ) ;
			
			// combine red-orange and red-violet to red
			cv::Mat red ;
			red = orange | violet ;
			
			// blur the image with a Gaussian filter to find a distinct maximum in each row
			cv::GaussianBlur( red, lineness, cv::Size( 25, 1 ), 0 ) ; // STUD_FUNCTION
			
			// exit case "linenessModeThreshold"
			break ;
		}
		
		// mode B: fuzzy logic
		case linenessModeFuzzy : {
			
			// limits where fuzzyness ends
			const double hRange = 90.0 ; // slice in color circle which may count as red, in [°]
			const double sMin   = 50.0 ; // minimum saturation limit in [0, 255]
			const double vMin   = 75.0 ; // STUD_SET_NAN minimum brightness value limit in range [0, 255]
			
			// divide hue range by 2*2 because one half is used for red-orange, the other for 
			// red-violet, and because the unit of hue values in the image is not [°] but [2°]
			const double hLimit = hRange / 4.0 ;
			
			// go through all image rows
			for( int row = 0 ; row < lineness.rows ; ++row ) {
				
				// go through all columns of the current row
				for( int col = 0 ; col < lineness.cols ; ++col ) {
					
					// get the current pixel at row "row" and column "col"
					const HSV8 hsvPix = hsv.at< HSV8 >( row, col ) ;
					
					// lineness factor, to be reduced by the criteria below
					double factor = 1.0 ;
					
					// reduce factor if hue is not red 
					     if( hsvPix.h < hLimit       ) { factor = factor * ( hLimit - hsvPix.h           ) / hLimit ; } // red-orange half
					else if( hsvPix.h > 180 - hLimit ) { factor = factor * ( hLimit - ( 180 - hsvPix.h ) ) / hLimit ; } // red-violet half
					else                               { factor = 0.0f ; }
					
					// reduce factor if saturation is low
					if( hsvPix.s > sMin ) { factor = factor * ( hsvPix.s - sMin ) / ( 255 - sMin ) ; }
					else                  { factor = STUD_NAN( 0.0 ) ; } // STUD_TODO what if saturation below limit?
					
					// reduce factor if brightness value is low
					if( hsvPix.v > vMin ) { factor = factor * STUD_NAN( hsvPix.v - vMin ) / ( 255 - vMin ) ; } // STUD_TODO what if brightness value above limit?
					else                  { factor = 0.0 ; }
					
					// apply factor to this pixel's lineness
					lineness.at< uint8_t >( row, col ) = 255 * factor ;
				}
			}
			
			// exit case "linenessModeFuzzy"
			break ;
		}
		
	} // switch "lineness modes"
	
	// only in debug mode, show lineness image and wait for a keystroke
	#ifdef DEBUG
		cv::imshow( "lineness", lineness ) ;
		cv::waitKey() ;
	#endif
	
	// 2) find most likely line point of each image row
	// ================================================
	
	// process one image row at a time
	std::vector< Point > linePoints ;
	for( int v = 0 ; v < lineness.rows ; ++v ) {
		
		// extract current row
		const std::vector< uint8_t > row = lineness.row( v ) ;
		
		// find row's pixel with maximum brightness/lineness
		std::vector< uint8_t >::const_iterator max = std::max_element( row.begin(), row.end() ) ; // STUD_SET_NOTHING
		
		// only process it if its lineness is big enough
		if( *max > 0 ) {
			
			// find "u" coordinate as distance between first pixel and maximum
			const int u = std::distance( row.begin(), max ) ; // STUD_SET_MINUSONE
			
			// add point to vector of found line points
			linePoints.push_back( Point( u, v ) ) ;
		}
	}
	
	// only in debug mode, print intermediate result
	#ifdef DEBUG
		hawaii::cout << "INFO: found possible line points in " << linePoints.size() << " of " << image.rows << " image rows" << hawaii::endl ;
	#endif
	
	// draw the found line points
	drawLinePoints( linePoints, visu ) ;
	
	// only in debug mode, show drawn line points and wait for a keystroke
	#ifdef DEBUG
		cv::imshow( "line detection", visu ) ;
		cv::waitKey() ;
	#endif
	
	// give up if not enough points have been found or if the first/last line point is not within the images upper/lower 
	// half, show drawn line points before returning
	if( (int)linePoints.size() < image.rows / 8 ) {
		hawaii::cerr << "ERROR: not enough red line points to continue or no points in upper or lower half of image" << hawaii::endl ;
		cv::imshow( "line detection", visu ) ;
		cv::waitKey( 1 ) ;
		return Coeffs() ;
	}
	
	// 3) find coefficients of polynomial line model
	// =============================================
	
	// three estimation modes available
	enum {
		estimatorLeastSquares,
		estimatorM,
		estimatorRANSAC
	} estimator = estimatorLeastSquares ; // STUD_TODO least-squares is enough for MTP, try all three for RVMRT
	
	// create measurements and measurement matrix with the correct dimensions
	cv::Mat_< double > y( linePoints.size(), 1 ) ;
	cv::Mat_< double > H( linePoints.size(), STUD_MINUSONE( 4 ) ) ; // STUD_TODO how many columns for 3rd order polynomial?
	
	// go through all found line points
	for( size_t idx = 0 ; idx < linePoints.size() ; ++idx ) {
		
		// get current point
		Point p = linePoints[ idx ] ;
		
		// set measurement for current point
		y( idx, 0 ) = p.u ;             // STUD_SET_NAN
		
		// set measurement matrix row for current point
		H( idx, 0 ) = p.v * p.v * p.v ; // STUD_SET_NAN
		H( idx, 1 ) = p.v * p.v       ; // STUD_SET_NAN
		H( idx, 2 ) = p.v             ; // STUD_SET_NAN
		H( idx, 3 ) = 1               ; // STUD_SET_NAN
	}
	
	// compute polynomial model coefficients
	cv::Mat_< double > p ;
	switch( estimator ) {
		
		// mode A: least squares
        case estimatorLeastSquares : {
			
			// use least-squares equation to estimate coefficients
			p = ( H.t() * H ).inv() * H.t() * y ; // STUD_SET_NAN
			
			// exit "switch" block
			break ;
		}
		
		// mode B: M-estimator
		case estimatorM: {
			
			// maximum acceptable error in [px] for "Huber" weight function
			const double errorTolerance = 10.0 ; // STUD_SET_NAN
			
			// initialize coefficients like with least squares
			p = ( H.t() * H ).inv() * H.t() * y ;
			
			// create error vector with the correct dimensions
			cv::Mat_< double > e( STUD_MINUSONE( y.rows ), 1 ) ; // STUD_TODO
			
			// create weight matrix with the correct dimensions, initialize as identity matrix ("Einheitsmatrix")
			cv::Mat_< double > W = cv::Mat_< double >::eye( H.rows, H.rows ) ;
			
			// buffer for previously estimated coefficients
			cv::Mat_< double > pPrev ; 
			
			// iteratively improve estimated coefficients, count the number of iterations
			size_t iter = 1 ;
			do {
				
				// draw the previos iteration's weighted line points and estimated model on a copy of the input image, then wait for a fixed time
				#ifdef DEBUG
				hawaii::cout << "INFO: M-Estimator iteration #" << iter << hawaii::endl ;
				cv::Mat visuMEst ;
				image.copyTo( visuMEst ) ;
				drawWeightedLinePoints( linePoints, W, visuMEst ) ;
				drawLineModel( p, visuMEst ) ;
				cv::imshow( "line detection (M-Estimator)", visuMEst ) ;
				cv::waitKey( 1000 ) ;
				#endif
				
				// compute error vector for current coefficients
				e = H * p - y ; // STUD_SET_NAN
				
				// adapt weights according to errors
				for( int idx = 0 ; idx < e.rows ; idx++ ) {
					
					// find absolute value of error in current matrix row
					const double errorAbs = std::abs( e( idx, 0 ) ) ;
					
					// implement Huber weight function: use full weight if error is within tolerance, reduce it otherwise
					if( errorAbs < errorTolerance ) {
						W( idx, idx ) = 2.0 ;
					} else {
						W( idx, idx ) = 2.0 * STUD_NAN( errorTolerance / errorAbs ) ; // STUD_TODO
					}
				}
				
				// remember previous polynomial coefficients 
				p.copyTo( pPrev ) ;
				
				// estimate new coefficients by including the weight matrix in the least-squares equation
				p = ( H.t() * W * H ).inv() * H.t() * W * y ; // STUD_SET_NAN
				
				// count this iteration
				++iter ;
				
			// stop after a given number of iterations or if coefficients do not change much any more 
			} while ( iter <= 50 && ( p(0,0) / pPrev(0,0) < 0.9 || p(0,0) / pPrev(0,0) > 1.1
			                       || p(1,0) / pPrev(1,0) < 0.9 || p(1,0) / pPrev(1,0) > 1.1
			                       || p(2,0) / pPrev(2,0) < 0.9 || p(2,0) / pPrev(2,0) > 1.1
			                       || p(3,0) / pPrev(3,0) < 0.9 || p(3,0) / pPrev(3,0) > 1.1 ) ) ;
			
			// only in debug mode, print intermediate result
			#ifdef DEBUG
				hawaii::cout << "INFO: ended M-estimator after " << iter-1 << " iterations" << hawaii::endl ;
			#endif
			
			// exit "switch" block
			break ;
		}
		
		// mode C: RANSAC
		case estimatorRANSAC: {
			
			// maximum acceptable error in [px] for counting a point as a supporter
			const double errorTolerance = 10.0 ; // STUD_SET_NAN
			
			// compute the required number of iterations
			const double probability = 0.9999 ; // required probability of choosing at least one sub-set without outliers
			const double ratio = 0.40 ; // maximum ratio of outliers in input line points 
			const double s = 4 ; // STUD_SET_NAN minimum number of points required for finding a 3rd order polynomial
			const size_t iters = log( 1 - probability ) / ( log( 1 - pow( 1 - ratio, s ) ) ) ; // STUD_SET_NAN
			
			// only in debug mode, print intermediate result
			#ifdef DEBUG
				hawaii::cout << "INFO: RANSAC will test " << iters << " sub-sets of " << s << " line points each" << hawaii::endl ;
			#endif
			
			// initialize random number generator based on current time
			srand( time( NULL ) ) ;
			
			// iteratively try to find a sub-set with the greatest number of supporters
			std::vector< size_t > supportersMax ;
			for( size_t iter = 1 ; iter <= iters ; ++iter ) {
				
				// generate 4 different random indices within the range of found line points
				const size_t range = linePoints.size() ;
				std::vector< size_t > random( 4 ) ;
				     random[ 0 ] = rand() % range ;
				do { random[ 1 ] = rand() % range ; } while( random[ 1 ] == random[ 0 ] ) ;
				do { random[ 2 ] = rand() % range ; } while( random[ 2 ] == random[ 0 ]
				                                          || random[ 2 ] == random[ 1 ] ) ;
				do { random[ 3 ] = rand() % range ; } while( random[ 3 ] == random[ 0 ]
				                                          || random[ 3 ] == random[ 1 ]
				                                          || random[ 3 ] == random[ 2 ] ) ;
				
				// extract sub-measurements and sub-measurement matrix with only these indices
				cv::Mat_< double > y4( 4, 1 ) ;
				cv::Mat_< double > H4( 4, 4 ) ;
				for( size_t i = 0 ; i < 4 ; ++i ) {
					y.row( random[ i ] ).copyTo( y4.row( i ) ) ;
					H.row( random[ i ] ).copyTo( H4.row( i ) ) ;
				}
				
				// find 3rd order polynomial exactly through the 4 random points
				cv::Mat_< double > p4 = ( H4.t() * H4 ).inv() * H4.t() * y4 ;
				
				// convert them to own convenient format
				Coeffs coeffs4 ;
				coeffs4[ 0 ] = p4( 0, 0 ) ;
				coeffs4[ 1 ] = p4( 1, 0 ) ;
				coeffs4[ 2 ] = p4( 2, 0 ) ;
				coeffs4[ 3 ] = p4( 3, 0 ) ;
				
				// count the total number of supporter points
				std::vector< size_t > supporters ;
				for( size_t idx = 0 ; idx < linePoints.size() ; ++idx ) {
					
					// get current line point
					const Point p = linePoints[ idx ] ;
					
					// find absolute error between point and polynomial model
					const double errorAbs = std::abs( coeffs4.u( p.v ) - p.u ) ; // STUD_SET_NAN
					
					// add its index to the supporter list if the error is small enough
					if( errorAbs < errorTolerance ) {
						supporters.push_back( idx ) ;
					}
				}
				
				// replace the old supporter list if this one is longer
				const bool betterSupporterSet = supporters.size() > supportersMax.size() ; // STUD_SET_ZERO
				if( betterSupporterSet ) {
					supportersMax = supporters ;
					
					// draw the so far best seed points, inliers and estimated model on a copy of the input image
					#ifdef DEBUG
					cv::Mat visuRANSACMax ;
					image.copyTo( visuRANSACMax ) ;
					drawRANSACLinePoints( linePoints, random, supportersMax, visuRANSACMax ) ;
					drawLineModel( p4, visuRANSACMax ) ;
					cv::imshow( "line detection (RANSAC so far best)", visuRANSACMax ) ;
					#endif
					
				}
				
				// draw the current seed points, inliers and estimated model on a copy of the input image, then wait for a fixed time
				#ifdef DEBUG
				hawaii::cout << "INFO: RANSAC iteration #" << iter << hawaii::endl ;
				cv::Mat visuRANSAC ;
				image.copyTo( visuRANSAC ) ;
				drawRANSACLinePoints( linePoints, random, supporters, visuRANSAC ) ;
				drawLineModel( p4, visuRANSAC ) ;
				cv::imshow( "line detection (RANSAC current)", visuRANSAC ) ;
				cv::waitKey( 200 ) ;
				#endif
			}
			
			// extract sub-measurements and sub-measurement matrix with all supporters
			cv::Mat_< double > yRANSAC( supportersMax.size(), 1 ) ;
			cv::Mat_< double > HRANSAC( supportersMax.size(), 4 ) ;
			for( size_t idx = 0; idx < supportersMax.size() ; ++idx ) {
				y.row( supportersMax[ idx ] ).copyTo( yRANSAC.row( idx ) ) ;
				H.row( supportersMax[ idx ] ).copyTo( HRANSAC.row( idx ) ) ;
			}
			
			// perform a standard least squares estimate based on the supporters only 
			p = ( HRANSAC.t() * HRANSAC ).inv() * HRANSAC.t() * yRANSAC ;
			
			// exit "switch" block
			break ;
		}
		
	} // switch "estimator modes"

	// draw the estimated polynomial and show the result image
	drawLineModel( p, visu ) ;
	cv::imshow( "line detection", visu ) ;
	cv::waitKey( 1 ) ;
	
	// convert estimated coefficients to own format and return them
	Coeffs coeffs ;
	coeffs[ 0 ] = p( 0, 0 ) ;
	coeffs[ 1 ] = p( 1, 0 ) ;
	coeffs[ 2 ] = p( 2, 0 ) ;
	coeffs[ 3 ] = p( 3, 0 ) ;
	return coeffs ;
}

// helper to draw line points in orange
void drawLinePoints( const std::vector< Point >& linePoints,
                           cv::Mat&              visu       ) {
	static const BGR8 orange( { 0, 127, 255 } ) ;
	for( const Point p : linePoints ) {
		if( p.v > 0 && p.v < visu.rows - 1
		 && p.u > 0 && p.u < visu.cols - 1 ) {
			visu.at< BGR8 >( p.v - 1, p.u     ) = orange ;
			visu.at< BGR8 >( p.v,     p.u - 1 ) = orange ;
			visu.at< BGR8 >( p.v,     p.u     ) = orange ;
			visu.at< BGR8 >( p.v,     p.u + 1 ) = orange ;
			visu.at< BGR8 >( p.v + 1, p.u     ) = orange ;
		}
	}
}

// helper to draw line points in orange, with opacity proportional to weight
void drawWeightedLinePoints( const std::vector< Point >& linePoints,
                             const cv::Mat_< double >    weights,
                                   cv::Mat&              visu       ) {
	HAWAII_ERROR_CONDITIONAL( linePoints.size() != (size_t)weights.rows
	                       || linePoints.size() != (size_t)weights.cols,
	                          "inconsistent number of points and weight dimensions" ) ;
	static const BGR8 orange( { 0, 127, 255 } ) ;
	for( size_t idx = 0 ; idx < linePoints.size() ; ++idx ) {
		const Point p = linePoints[ idx ] ;
		const unsigned int w = std::max( 0u, std::min( 256u, (unsigned int)( 128.0 * weights( idx, idx ) ) ) ) ;
		if( p.v > 0 && p.v < visu.rows - 1
		 && p.u > 0 && p.u < visu.cols - 1 ) {
			visu.at< BGR8 >( p.v - 1, p.u     ).b = ( w * orange.b + ( 256-w ) * visu.at< BGR8 >( p.v - 1, p.u     ).b ) >> 8 ;
			visu.at< BGR8 >( p.v - 1, p.u     ).g = ( w * orange.g + ( 256-w ) * visu.at< BGR8 >( p.v - 1, p.u     ).g ) >> 8 ;
			visu.at< BGR8 >( p.v - 1, p.u     ).r = ( w * orange.r + ( 256-w ) * visu.at< BGR8 >( p.v - 1, p.u     ).r ) >> 8 ;
			visu.at< BGR8 >( p.v,     p.u - 1 ).b = ( w * orange.b + ( 256-w ) * visu.at< BGR8 >( p.v,     p.u - 1 ).b ) >> 8 ;
			visu.at< BGR8 >( p.v,     p.u - 1 ).g = ( w * orange.g + ( 256-w ) * visu.at< BGR8 >( p.v,     p.u - 1 ).g ) >> 8 ;
			visu.at< BGR8 >( p.v,     p.u - 1 ).r = ( w * orange.r + ( 256-w ) * visu.at< BGR8 >( p.v,     p.u - 1 ).r ) >> 8 ;
			visu.at< BGR8 >( p.v,     p.u     ).b = ( w * orange.b + ( 256-w ) * visu.at< BGR8 >( p.v,     p.u     ).b ) >> 8 ;
			visu.at< BGR8 >( p.v,     p.u     ).g = ( w * orange.g + ( 256-w ) * visu.at< BGR8 >( p.v,     p.u     ).g ) >> 8 ;
			visu.at< BGR8 >( p.v,     p.u     ).r = ( w * orange.r + ( 256-w ) * visu.at< BGR8 >( p.v,     p.u     ).r ) >> 8 ;
			visu.at< BGR8 >( p.v,     p.u + 1 ).b = ( w * orange.b + ( 256-w ) * visu.at< BGR8 >( p.v,     p.u + 1 ).b ) >> 8 ;
			visu.at< BGR8 >( p.v,     p.u + 1 ).g = ( w * orange.g + ( 256-w ) * visu.at< BGR8 >( p.v,     p.u + 1 ).g ) >> 8 ;
			visu.at< BGR8 >( p.v,     p.u + 1 ).r = ( w * orange.r + ( 256-w ) * visu.at< BGR8 >( p.v,     p.u + 1 ).r ) >> 8 ;
			visu.at< BGR8 >( p.v + 1, p.u     ).b = ( w * orange.b + ( 256-w ) * visu.at< BGR8 >( p.v + 1, p.u     ).b ) >> 8 ;
			visu.at< BGR8 >( p.v + 1, p.u     ).g = ( w * orange.g + ( 256-w ) * visu.at< BGR8 >( p.v + 1, p.u     ).g ) >> 8 ;
			visu.at< BGR8 >( p.v + 1, p.u     ).r = ( w * orange.r + ( 256-w ) * visu.at< BGR8 >( p.v + 1, p.u     ).r ) >> 8 ;
		}
	}
}

// helper to draw RANSAC seed points as big yellow X and supporters in orange
void drawRANSACLinePoints( const std::vector< Point >&  linePoints,
                           const std::vector< size_t >& seeds,
                           const std::vector< size_t >& supporters,
                                 cv::Mat&               visu       ) {
	static const BGR8 orange( { 0, 127, 255 } ) ;
	static const BGR8 yellow( { 0, 255, 255 } ) ;
	for( const auto idx : supporters ) {
		const Point p = linePoints[ idx ] ;
		if( p.v > 0 && p.v < visu.rows - 1
		 && p.u > 0 && p.u < visu.cols - 1 ) {
			visu.at< BGR8 >( p.v - 1, p.u     ) = orange ;
			visu.at< BGR8 >( p.v,     p.u - 1 ) = orange ;
			visu.at< BGR8 >( p.v,     p.u     ) = orange ;
			visu.at< BGR8 >( p.v,     p.u + 1 ) = orange ;
			visu.at< BGR8 >( p.v + 1, p.u     ) = orange ;
		}
	}
	for( const auto idx : seeds ) {
		const Point p = linePoints[ idx ] ;
		if( p.v > 1 && p.v < visu.rows - 2
		 && p.u > 1 && p.u < visu.cols - 2 ) {
			visu.at< BGR8 >( p.v - 2, p.u - 2 ) = yellow ;
			visu.at< BGR8 >( p.v - 2, p.u - 1 ) = yellow ;
			visu.at< BGR8 >( p.v - 2, p.u + 1 ) = yellow ;
			visu.at< BGR8 >( p.v - 2, p.u + 2 ) = yellow ;
			visu.at< BGR8 >( p.v - 1, p.u - 2 ) = yellow ;
			visu.at< BGR8 >( p.v - 1, p.u - 1 ) = yellow ;
			visu.at< BGR8 >( p.v - 1, p.u     ) = yellow ;
			visu.at< BGR8 >( p.v - 1, p.u + 1 ) = yellow ;
			visu.at< BGR8 >( p.v - 1, p.u + 2 ) = yellow ;
			visu.at< BGR8 >( p.v,     p.u - 1 ) = yellow ;
			visu.at< BGR8 >( p.v,     p.u     ) = yellow ;
			visu.at< BGR8 >( p.v,     p.u + 1 ) = yellow ;
			visu.at< BGR8 >( p.v + 1, p.u - 2 ) = yellow ;
			visu.at< BGR8 >( p.v + 1, p.u - 1 ) = yellow ;
			visu.at< BGR8 >( p.v + 1, p.u     ) = yellow ;
			visu.at< BGR8 >( p.v + 1, p.u + 1 ) = yellow ;
			visu.at< BGR8 >( p.v + 1, p.u + 2 ) = yellow ;
			visu.at< BGR8 >( p.v + 2, p.u - 2 ) = yellow ;
			visu.at< BGR8 >( p.v + 2, p.u - 1 ) = yellow ;
			visu.at< BGR8 >( p.v + 2, p.u + 1 ) = yellow ;
			visu.at< BGR8 >( p.v + 2, p.u + 2 ) = yellow ;
		}
	}
}

// helper to draw a polynomial in yellow
void drawLineModel( const cv::Mat_< double > coeffs,
                          cv::Mat&           visu   ) {
	HAWAII_ERROR_CONDITIONAL( coeffs.rows != 4
	                       || coeffs.cols != 1,
	                          "invalid coefficient dimensions" ) ;
	static const BGR8 yellow( { 0, 255, 255 } ) ;
	for( int v = 0 ; v < visu.rows ; ++v ) {
		const int uLine = std::lround( coeffs( 0, 0 ) * v * v * v
		                             + coeffs( 1, 0 ) * v * v
		                             + coeffs( 2, 0 ) * v
		                             + coeffs( 3, 0 )             ) ;
		for( int u = std::max( uLine - 1, 0         ) ;
		         u < std::min( uLine + 1, visu.cols ) ;
		       ++u                                   ) {
			visu.at< BGR8 >( v, u ) = yellow ;
		}
	}
}
